/**
  *  \file game/expl.cc
  *  \brief Explosions
  *
  *  In PCC2, explosions are a separate type. In PCC 1.x, they were
  *  markers with a special tag. Implementing them as separate type
  *  allows to attach regular information in a more meaningful way
  *  than by recycling and abusing drawings. The disadvantage is that
  *  we need to process two info sources when iterating over markers
  *  and explosions. This will mainly become an issue when we want to
  *  provide a uniform interface for scripts.
  *
  *  \todo make GExplosion and GExplosionContainer implement GObject
  *  and GObjectType, respectively.
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/expl.h"

/** Create explosion.
    \param id  Explosion Id (NOT ship Id!), 0 if not known.
    \param pos Explosion position. */
GExplosion::GExplosion(int id, GPoint pos)
    : id(id), pos(pos), ship_name(), ship_id(0)
{ }

GExplosion::~GExplosion()
{ }

/** Get name of ship that exploded here. Empty if unknown. */
string_t
GExplosion::getShipName() const
{
    return ship_name;
}

/** Set name of ship that exploded here. */
void
GExplosion::setShipName(string_t name)
{
    this->ship_name = name;
}

/** Merge information of other explosion record. This tests whether
    these records potentially describe the same explosion and, if yes,
    merges them.
    \param other other explosion record
    \return true if merge successful */
bool
GExplosion::merge(const GExplosion& other)
{
    /* not same position? Cannot match. */
    if (pos != other.pos)
        return false;

    /* different explosion Id? */
    if (id != 0 && other.id != 0 && id != other.id)
        return false;

    /* different ship name? */
    // FIXME: deal with dummy ship names produced by AllowShipNames=No
    if (ship_name.size() && other.ship_name.size() && ship_name != other.ship_name)
        return false;

    /* different id? */
    if (ship_id && other.ship_id && ship_id != other.ship_id)
        return false;

    /* ok, it will work. do it. */
    if (other.id) {
        id = other.id;
    }
    if (other.ship_name.size()) {
        ship_name = other.ship_name;
        ship_id   = other.ship_id;
    } else if (other.ship_id) {
        ship_id   = other.ship_id;
    }
    return true;
}

/************************** GExplosionContainer **************************/

/** Constructor. */
GExplosionContainer::GExplosionContainer()
{ }

/** Destructor. */
GExplosionContainer::~GExplosionContainer()
{ }

/** Add explosion. If this explosion matches one we already know,
    merges the information.
    \param x Explosion to add
    \return reference to added/modified explosion record */
GExplosion&
GExplosionContainer::addExplosion(GExplosion x)
{
    for (iterator i = begin(); i != end(); ++i) {
        if (i->merge(x)) {
            return *i;
        }
    }
    explosions.push_back(x);
    return explosions.back();
}
