/**
  *  \file game/msgtemplate.cc
  *  \brief Message Template
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/msgtemplate.h"
#include "util/console.h"
#include "game/racename.h"
#include "game/struct.h"
#include "util/misc.h"
#include "game/spec.h"
#include "game/player.h"
#include "cpluslib/format.h"


/** Format a number into a string.
    We cannot use numToString, because that honors user configuration
    and might mangle the values to be not parseable. */
static string_t
formatNumber(int i)
{
    return format("%d", i);
}

/** Parse integer value according to a type modifier.
    \param value Value from message
    \param type Type modifier from template
    \return transformed string (usually, a stringified integer)

    \todo Implement sloppy compare. Text from messages can have Unicode
    characters replaced by spaces by host. It might also have been
    truncated. The same goes for hull names.

    \todo It would make sense to detect when the player's files do not
    match host's. However, we must be really sure about this: in a game
    with fewer than 11 players, the FreeFighters template will try to
    interpret config strings like "max mine radius 150" as name/count
    pairs. */
static string_t
prepareValue(string_t value, string_t type)
{
    if (type.size() > 7 && type.compare(type.size()-7, 7, "+ALLIES", 7) == 0) {
        /* Value includes an alliance marker, but we just want the plain name.
           The message line reads "Bird Man ! :    0".
             !     => this race has offered something to us
             +     => we have offered something
             :     => since HOST sometimes drops the colon, we tell the matching engine
                      to include it in the name, which means we must strip it here
           Parsing of the actual alliances is done in msgparse.cc, generateFlagAllies(). */
        string_t::size_type n = value.size();
        while (n > 0 && (value[n-1] == '+' || value[n-1] == '!' || value[n-1] == ':' || value[n-1] == ' '))
            --n;
        value.erase(n);
        type.erase(type.size()-7);
    }

    value = strTrim(value);
    if (type == "RACE") {
        /* Full race name */
        for (int i = 1; i <= NUM_OWNERS; ++i)
            if (strCaseCmp(value, strTrim(host_racenames.getFullName(i))) == 0)
                return formatNumber(i);
        return string_t();
    } else if (type == "RACE.SHORT") {
        /* Short race name */
        for (int i = 1; i <= NUM_OWNERS; ++i)
            if (strCaseCmp(value, strTrim(host_racenames.getShortName(i))) == 0)
                return formatNumber(i);
        return string_t();
    } else if (type == "RACE.ADJ") {
        /* Race adjective name */
        for (int i = 1; i <= NUM_OWNERS; ++i)
            if (strCaseCmp(value, strTrim(host_racenames.getAdjName(i))) == 0)
                return formatNumber(i);
        return string_t();
    } else if (type == "HULL") {
        /* Hull name */
        for (int i = 1; i <= getNumHulls(); ++i)
            if (strCaseCmp(value, strTrim(getHull(i).getName())) == 0)
                return formatNumber(i);
        return string_t();
    } else if (type == "X10" || type == "X100" || type == "X1000") {
        /* Scaled integer */
        string_t::size_type start = (value.empty() || (value[0] != '+' && value[0] != '-')) ? 0 : 1;
        string_t::size_type n = value.find_first_not_of("0123456789", start);
        if (n == string_t::npos) {
            /* completely numeric */
            return value + type.substr(2);
        } else if (value[n] == '.') {
            /* includes decimal point */
            value.erase(n, 1);
            string_t::size_type i = 0;
            while (n+i < value.size() && i+2 < type.size() && value[n+i] >= '0' && value[n+i] <= '9')
                ++i;
            value.erase(n+i);
            while (i+2 < type.size())
                value += '0', ++i;
            return value;
        } else {
            /* no decimal point */
            return value.substr(0, n) + type.substr(2);
        }
    } else {
        string_t::size_type n = type.find('/');
        if (n != type.npos) {
            /* Enum */
            string_t::size_type p = 0;
            int i = 0;
            while (n != type.npos) {
                if (strCaseCmp(value, strTrim(string_t(type, p, n-p))) == 0)
                    return formatNumber(i);
                ++i;
                p = n+1;
                n = type.find('/', p);
            }
            if (strCaseCmp(value, strTrim(string_t(type, p))) == 0)
                return formatNumber(i);
            return string_t();
        } else {
            /* Unparsed */
            return value;
        }
    }
}

/** Convert string to integer. */
static int32_t
parseIntValue(string_t value)
{
    string_t::size_type pos;
    long result;
    if (stringValue(value, result, pos) || stringValue(value.substr(0, pos), result))
        return result;
    else
        return -1;
}

/** Split message into lines. */
void
splitMessage(GMessageLines& out, const string_t& in)
{
    string_t::size_type i = 0;
    for (string_t::size_type n = in.find('\n'); n != in.npos; i = n+1, n = in.find('\n', i))
        out.push_back(string_t(in, i, n-i));
    out.push_back(string_t(in, i));
}

/**************************** GMessageTemplate ***************************/

/** Constructor. Create a blank message template.
    \param message_object object kind described by this template
    \param name name of the template, for reference by user */
GMessageTemplate::GMessageTemplate(GMessageObject message_object, string_t name)
    : message_object(message_object),
      name(name),
      continue_flag(false)
{ }

GMessageTemplate::~GMessageTemplate()
{ }

/** Add a iMatchKind, iMatchSubId, iMatchBigId instruction.
    \param opcode Opcode
    \param value  Value to match */
void
GMessageTemplate::addMatchInstruction(uint8_t opcode, uint16_t value)
{
    instructions.push_back(Instruction(opcode, 0, value, 0));
}

/** Add iValue instructions.
    \param opcode should be iValue
    \param value  List of values to produce, separated by commas. Each entry will generate one instruction.  */
void
GMessageTemplate::addValueInstruction(uint8_t opcode, string_t value)
{
    string_t::size_type p = 0, n;
    for (n = value.find(','); n != string_t::npos; p = n+1, n = value.find(',', p)) {
        instructions.push_back(Instruction(opcode, 0, strings.size(), 0));
        strings.push_back(strTrim(string_t(value, p, n-p)));
    }
    instructions.push_back(Instruction(opcode, 0, strings.size(), 0));
    strings.push_back(strTrim(string_t(value, p)));
}

/** Add a check/parse instruction.
    \param opcode Opcode, iCheck to iArray, with optional scope modifier added.
    \param offset Distance, according to scope modifier.
    \param value  Value to match. For iParse/iArray, wildcard symbols ('$') are processed. */
void
GMessageTemplate::addCheckInstruction(uint8_t opcode, int8_t offset, string_t value)
{
    instructions.push_back(Instruction(opcode, offset, strings.size(), 0));
    if ((opcode & iMask) == iParse || (opcode & iMask) == iArray) {
        string_t::size_type p = 0, n;
        for (n = value.find('$'); n != string_t::npos; p = n+1, n = value.find('$', p)) {
            strings.push_back(string_t(value, p, n-p));
            ++instructions.back().count;
        }
        strings.push_back(string_t(value, p));
    } else {
        strings.push_back(value);
    }
}

/** Add single variable.
    \param name name or name:type pair */
void
GMessageTemplate::addVariable(string_t name)
{
    string_t type;
    strSplit(name, name, type, ":");
    variables.push_back(strTrim(strUCase(name)));
    types.push_back(strTrim(strUCase(type)));
}

/** Add variables.
    \param names comma-separated list of variable names */
void
GMessageTemplate::addVariables(string_t names)
{
    string_t::size_type p = 0, n;
    for (n = names.find(','); n != string_t::npos; p = n+1, n = names.find(',', p)) {
        addVariable(string_t(names, p, n-p));
    }
    addVariable(string_t(names, p));
}

/** Set continue flag. If set, matching this template does not stop matching further templates.
    \param flag Value to set */
void
GMessageTemplate::setContinueFlag(bool flag)
{
    this->continue_flag = flag;
}

/** Get number of variables. This is the number of variables in "assign" statements that this
    template will generate. */
int
GMessageTemplate::getNumVariables() const
{
    return variables.size();
}

/** Get number of wildcards. This is the number of values this template will produce when
    matching. */
int
GMessageTemplate::getNumWildcards() const
{
    int total = 0;
    for (std::vector<Instruction>::const_iterator i = instructions.begin(); i != instructions.end(); ++i) {
        if (i->opcode == iValue || (i->opcode & iMask) == iFind) {
            ++total;
        } else if ((i->opcode & iMask) == iParse || (i->opcode & iMask) == iArray) {
            total += i->count;
        }
    }
    return total;
}

/** Get number of restrictions. This is the number of non-empty matches. A template with no
    restrictions will match every message. */
int
GMessageTemplate::getNumRestrictions() const
{
    int total = 0;
    for (std::vector<Instruction>::const_iterator i = instructions.begin(); i != instructions.end(); ++i) {
        if (i->opcode != iValue) {
            ++total;
        }
    }
    return total;
}

/** Find variable slot by name.
    \return index such that getVariableName(return) == name, -1 if none. */
int
GMessageTemplate::getVariableSlotByName(const string_t name) const
{
    for (uint_t i = 0; i < variables.size(); ++i)
        if (name == variables[i])
            return i;
    return -1;
}

/** Get variable name.
    \param index Variable index [0,getNumVariables()) */
string_t
GMessageTemplate::getVariableName(int index) const
{
    if (index < 0 || index >= getNumVariables())
        return string_t();
    else
        return variables[index];
}

/** Get template name. */
string_t
GMessageTemplate::getTemplateName() const
{
    return name;
}


/** Match message against this template.
    \param message [in] Message
    \param player  [in] Player who got the message
    \param values  [out] Placeholder values; empty on call.
    \retval true if message matches, values have been filled out
    \retval false if message does not match, values has undefined content */
bool
GMessageTemplate::match(const GMessageLines& message, int player, std::vector<string_t>& values) const
{
    uint16_t line = 0;
    for (std::vector<Instruction>::const_iterator it = instructions.begin(); it != instructions.end(); ++it) {
        uint8_t opcode = it->opcode;
        uint8_t group  = opcode & iMask;
        if (opcode == iMatchKind) {
            /* Match MsgHdrKind */
            if (getMessageHeaderInformation(message, MsgHdrKind) != it->index)
                return false;
        } else if (opcode == iMatchSubId) {
            /* Match MsgHdrSubId */
            if (getMessageHeaderInformation(message, MsgHdrSubId) != it->index)
                return false;
        } else if (opcode == iMatchBigId) {
            /* Match MsgHdrBigId */
            if (getMessageHeaderInformation(message, MsgHdrBigId) != it->index)
                return false;
        } else if (opcode == iValue) {
            /* Unconditionally produce a value */
            const string_t& s = strings[it->index];
            if (strCaseCmp(s, "player") == 0) {
                /* Player has been given by the caller */
                values.push_back(formatNumber(player));
            } else if (strCaseCmp(s, "id") == 0) {
                /* Id is an integer */
                values.push_back(formatNumber(getMessageHeaderInformation(message, MsgHdrId)));
            } else if (strCaseCmp(s, "bigid") == 0) {
                /* BigId is an integer */
                values.push_back(formatNumber(getMessageHeaderInformation(message, MsgHdrBigId)));
            } else if (strCaseCmp(s, "subid") == 0) {
                /* The message header information is a character. This is intended
                   for the case where the character is a race number (0-9, a, b),
                   and we want an integer. SubId can also be c (for colonists),
                   which we do not handle here. */
                values.push_back(formatNumber(getPlayerIdFromChar(getMessageHeaderInformation(message, MsgHdrSubId))));
            } else {
                values.push_back(s);
            }
        } else if (group == iCheck) {
            /* Fail if string not found */
            if (!check(message, line, *it))
                return false;
        } else if (group == iFail) {
            /* Fail if string found */
            if (check(message, line, *it))
                return false;
        } else if (group == iFind) {
            /* Remember found status */
            if (check(message, line, *it))
                values.push_back("1");
            else
                values.push_back("0");
        } else if (group == iParse || group == iArray) {
            /* Parse or Array: start by locating the match */
            uint_t typeIndex = values.size();
            if ((opcode & ~iMask) == sAny) {
                for (GMessageLines::size_type i = 0; i < message.size(); ++i) {
                    if (matchLine(message[i], it->index, it->count, typeIndex, values)) {
                        line = i;
                        goto ok;
                    }
                }
                return false;
             ok:;
            } else {
                if ((opcode & ~iMask) == sRelative) {
                    line += it->offset;
                } else {
                    line = it->offset - 1;
                }
                if (line >= message.size())
                    return false;
                if (!matchLine(message[line], it->index, it->count, typeIndex, values))
                    return false;
            }

            /* If it's an array, read the additional lines */
            if (group == iArray) {
                uint_t nelems = 1;
                while (nelems < 11 && line+1U < message.size() && matchLine(message[line+1], it->index, it->count, typeIndex, values))
                    ++nelems, ++line;
                consolidateArray(values, it->count, nelems);
            }
        } else {
            return false;
        }
    }
    return true;
}

#if 1
/** Dump. For debugging use. */
void
GMessageTemplate::dump() const
{
    static const char*const scopes[] = { "any", "rel", "fix" };
    for (std::vector<Instruction>::const_iterator i = instructions.begin(); i != instructions.end(); ++i) {
        enum { Index = 0, String = 1, NStrings = 2, Pos = 16 };
        int arg;
        if (i->opcode == iMatchKind) {
            console << "matchkind"; arg = Index;
        } else if (i->opcode == iMatchSubId) {
            console << "matchsubid"; arg = Index;
        } else if (i->opcode == iMatchBigId) {
            console << "matchbigid"; arg = Index;
        } else if (i->opcode == iValue) {
            console << "value"; arg = String;
        } else if ((i->opcode & iMask) == iCheck) {
            console << "check"; arg = String + Pos;
        } else if ((i->opcode & iMask) == iFail) {
            console << "fail"; arg = String + Pos;
        } else if ((i->opcode & iMask) == iFind) {
            console << "find"; arg = String + Pos;
        } else if ((i->opcode & iMask) == iParse) {
            console << "parse"; arg = NStrings + Pos;
        } else if ((i->opcode & iMask) == iArray) {
            console << "array"; arg = NStrings + Pos;
        } else {
            console << "bug"; arg = Index;
        }

        if (arg & Pos) {
            console << scopes[i->opcode & ~iMask];
            arg &= ~Pos;
        }
        if (arg == Index) {
            console << " " << i->index << "\n";
        }
        if (arg == String) {
            console << " '" << strings[i->index] << "'\n";
        }
        if (arg == NStrings) {
            console << " '" << strings[i->index] << "'";
            for (uint_t j = 0; j < i->count; ++j)
                console << ",  '" << strings[i->index + j + 1] << "'";
            console << "\n";
        }
    }
}
#endif

/** Convert string value to integer. */
int32_t
GMessageTemplate::getIntValue(const string_t& value)
{
    // FIXME: do we need to parse booleans?
    return ::parseIntValue(value);
}

/** Check for a string.
    \param message [in] Message text
    \param line    [in/out] Current line pointer
    \param insn    [in] Instruction to process; this checks the index and scope fields */
bool
GMessageTemplate::check(const GMessageLines& message, uint16_t& line, const Instruction& insn) const
{
    string_t needle = strUCase(host_racenames.expandNames(strings[insn.index]));
    if ((insn.opcode & ~iMask) == sAny) {
        /* Check all lines */
        for (GMessageLines::size_type i = 0; i < message.size(); ++i) {
            if (strUCase(message[i]).find(needle) != string_t::npos) {
                /* Found it */
                line = i;
                return true;
            }
        }
        return false;
    } else {
        /* Check just one line */
        uint16_t lineToCheck;
        if ((insn.opcode & ~iMask) == sRelative) {
            lineToCheck = line + insn.offset;
        } else {
            lineToCheck = insn.offset - 1;    /* user gives position 1-based */
        }
        if (lineToCheck < message.size() && strUCase(message[lineToCheck]).find(needle) != string_t::npos) {
            line = lineToCheck;
            return true;
        } else {
            return false;
        }
    }
}

/** Match a single line.
    \param line   [in] Text line from message
    \param index  [in] Index into string table of first string
    \param nvar   [in] Number of variables to produce
    \param typeIndex [in] Index into type array
    \param values [out] Produce output here
    \retval true match succeeded; values have been produced and postprocessed according to their types
    \retval false match failed; values have not been changed */
bool
GMessageTemplate::matchLine(const string_t& line, uint_t index, uint_t nvar, uint_t typeIndex, std::vector<string_t>& values) const
{
    string_t needle = strUCase(host_racenames.expandNames(strings[index]));
    string_t::size_type pos = (needle.empty() ? 0 : strUCase(line).find(needle));
    if (pos != string_t::npos) {
        if (matchPart(line, pos + needle.size(), index+1, nvar, values)) {
            /* Postprocess values */
            uint_t valueIndex = values.size() - nvar;
            for (uint_t i = 0; i < nvar && typeIndex < types.size(); ++i, ++typeIndex, ++valueIndex) {
                values[valueIndex] = prepareValue(values[valueIndex], types[typeIndex]);
            }
            return true;
        } else {
            return false;
        }
    } else {
        return false;
    }
}

/** Match a partial line.
    \param line    [in] Text line from message
    \param startAt [in] Value of variable starts at this position
    \param index   [in] Index into string table of first string
    \param nvar    [in] Number of variables to produce
    \param values  [out] Produce output here
    \retval true match succeeded; values have been produced (but not postprocessed)
    \retval false match failed; values have not been changed */
bool
GMessageTemplate::matchPart(const string_t& line, uint_t startAt, uint_t index, uint_t nvar, std::vector<string_t>& values) const
{
    /* Special cases: */
    if (nvar == 0) {
        return true;
    }
    if (nvar == 1 && strings[index].empty()) {
        values.push_back(line.substr(startAt));
        return true;
    }

    /* Build initial hypothesis */
    string_t needle = strUCase(host_racenames.expandNames(strings[index]));
    string_t::size_type pos = strUCase(line).rfind(needle);
    if (pos <= startAt || pos == string_t::npos) {
        return false;
    }
    values.push_back(string_t(line, startAt, pos - startAt));

    /* Check subsequent patterns, backtrack if needed */
    while (values.back().find_first_not_of(' ') == string_t::npos
           || !matchPart(line, pos + needle.size(), index+1, nvar-1, values))
    {
        if (pos == 0) {
            /* We cannot backtrack */
            values.pop_back();
            return false;
        }

        string_t::size_type npos = strUCase(line).rfind(needle, pos-1);
        if (npos <= startAt || npos >= pos) {
            values.pop_back();
            return false;
        }
        values.back().assign(line, startAt, npos - startAt);
        pos = npos;
    }
    return true;
}

/** Consolidate an array. Matching has produced nvar*nelems items.
    Combine that down to nvar elements, each containing a list of values.
    \param values [in/out] Values
    \param nvar   [in] Number of variables (columns)
    \param nelems [in] Number of array items (rows) */
void
GMessageTemplate::consolidateArray(std::vector<string_t>& values, uint_t nvar, uint_t nelems) const
{
    const uint_t nil = uint_t(-1);
    uint_t firstIndex = values.size() - nvar*nelems;

    /* Figure out whether there's an element which defines the index. */
    uint_t playerIndex = nil;
    for (uint_t i = 0; i < nvar && i+firstIndex < variables.size(); ++i) {
        if (variables[i+firstIndex] == "INDEX") {
            playerIndex = i;
            break;
        }
    }

    /* Build all elements. We always build 11-element arrays. */
    for (uint_t var = 0; var < nvar; ++var) {
        /* This one's special */
        if (var == playerIndex)
            continue;

        /* Parse individual items. Unset items remain blank. */
        string_t data[NUM_PLAYERS];
        for (uint_t i = 0; i < nelems; ++i) {
            int slot;
            if (playerIndex == nil)
                slot = i;
            else
                slot = parseIntValue(values[playerIndex + i*nvar + firstIndex]) - 1;

            if (slot >= 0 && slot < NUM_PLAYERS)
                data[slot] = strTrim(values[var + i*nvar + firstIndex]);
        }

        /* Build new value */
        for (uint_t i = 1; i < NUM_PLAYERS; ++i) {
            data[0] += ',';
            data[0] += data[i];
        }
        values[var+firstIndex] = data[0];
    }

    /* Clear player index to avoid that anyone uses it */
    if (playerIndex != nil)
        values[playerIndex+firstIndex].clear();

    /* Remove excess values */
    values.resize(firstIndex + nvar);
}

/** Extract information from message header.
    \param msg  Message text
    \param what Requested information
    \return Information. 0 if not found. */
int
getMessageHeaderInformation(const GMessageLines& msg, GMessageHeaderInformation what)
{
    /* Must be at least one line, and at least 5 characters "(-x0)" */
    if (msg.empty() || msg[0].size() < 5)
        return 0;

    /* Check format */
    if (msg[0][0] != '(' || (msg[0][1] != '-' && msg[0][1] != 'o'))
        return 0;

    /* Kind, Age or SubId */
    if (what == MsgHdrAge)
        return msg[0][1] != '-';
    if (what == MsgHdrKind)
        return msg[0][2];
    if (what == MsgHdrSubId)
        return msg[0][3];

    /* Id or BigId */
    int result = 0;
    string_t::size_type start = (what == MsgHdrBigId ? 3 : 4);
    while (start < msg[0].size() && msg[0][start] != ')') {
        char ch = msg[0][start];
        if (ch >= '0' && ch <= '9')
            result = 10*result + ch - '0';
        ++start;
    }
    return result;
}
