/**
  *  \file game/player.cc
  *  \brief Player Routines
  *
  *  This module contains various routines that deal with players.
  *  - Global player numbers are now in game/team.cc.
  *  - Race names are in game/racename.cc.
  *  - Player sets are in game/playerset.cc.
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <string>
#include "game/player.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/struct.h"

/** List of player characters. The abbreviation for player N in a
    friendly code is player_chars[N]. */
static const char player_chars[] = "0123456789ABX";

/** Parse player character into player number. This is used for
    subspace messages, fcodes.cc, etc.
    \returns player id [0,11], or -1. */
int
getPlayerIdFromChar(char c)
{
    if (c >= '0' && c <= '9')
        return c - '0';
    else if (c == 'A' || c == 'a')
        return 10;
    else if (c == 'B' || c == 'b')
        return 11;
    else
        return -1;
}

/** Get character representing a player. */
char
getPlayerChar(int player)
{
    return player_chars[player];
}

/** Parse list of players from list of letters. Parses a "123456789ab"-style
    string up to its end or a ',', and leaves the parse position there.
    Used for parsing mission definitions etc.
    \param text [in] Text to parse
    \param index [in/out] Parse position
    \return player set */
GPlayerSet
parsePlayerList(const string_t& text, string_t::size_type& index)
{
    // FIXME: can we merge this with misc.cc:parseFlags?
    GPlayerSet result;
    while (index < text.size() && text[index] != ',') {
        int p = getPlayerIdFromChar(text[index]);
        if (p > 0)
            result += p;
        ++index;
    }
    return result;
}
