/**
  *  \file gfx/clip.cc
  *  \brief Simple Clipping Implementation
  *
  *  Classes for clipping output at a single rectangle.
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "gfx/clip.h"

/** \class GfxClipRect
    \brief Clip at a Rectangle

    This class provides the back-end for drawing with clipping
    at a rectangle.

    It does not contain a public constructor or methods to modify the
    clip rectangle; derived classes must implement these when needed. */

/** Constructor.
    \param rect      clipping rectangle
    \param parent    canvas to draw on */
GfxClipRect::GfxClipRect(const GfxRect& rect, GfxCanvas& parent)
    : GfxFilter(parent), rect(rect)
{ }

void
GfxClipRect::drawHLine(const GfxPoint& pt, int npix, uint32 color, uint8 pat, alpha_type alpha)
{
    GfxRect r(pt, npix, 1);
    r.clipRect(rect);
    if (r.exists())
        parent.drawHLine(GfxPoint(r.x, r.y), r.w, color, pat, alpha);
}

void
GfxClipRect::drawVLine(const GfxPoint& pt, int npix, uint32 color, uint8 pat, alpha_type alpha)
{
    GfxRect r(pt, 1, npix);
    r.clipRect(rect);
    if (r.exists())
        parent.drawVLine(GfxPoint(r.x, r.y), r.h, color, pat, alpha);
}

void
GfxClipRect::drawPixel(const GfxPoint& pt, uint32 color, alpha_type alpha)
{
    if (rect.contains(pt))
        parent.drawPixel(pt, color, alpha);
}

void
GfxClipRect::drawBar(GfxRect r, uint32 color, uint32 bg,
                     const GfxFillPattern& pat, alpha_type alpha)
{
    r.clipRect(rect);
    if (r.exists())
        parent.drawBar(r, color, bg, pat, alpha);
}

void
GfxClipRect::blitSurface(const GfxPoint& pt, GfxCanvas& src, GfxRect r)
{
    GfxRect rclip = rect;
    rclip.moveBy(-pt.x, -pt.y);
    r.clipRect(rclip);
    if (r.exists())
        parent.blitSurface(pt, src, r);
}

void
GfxClipRect::blitPattern(GfxRect r, const GfxPoint& pt, int bytes_per_line,
                         const uint8* data, uint32 color,
                         uint32 bg, alpha_type alpha)
{
    r.clipRect(rect);
    if (r.exists())
        parent.blitPattern(r, pt, bytes_per_line, data, color, bg, alpha);
}

GfxRect
GfxClipRect::computeClipRect(GfxRect r)
{
    r.clipRect(rect);
    return parent.computeClipRect(r);
}

bool
GfxClipRect::computeOffset(GfxCanvas& other, GfxPoint& pt)
{
    return this == &other
        || parent.computeOffset(other, pt);
}

/*************************** GfxClipRectFilter ***************************/

/** \class GfxClipRectFilter
    \brief Clipping at a Rectangle

    This class provides clipping at a rectangle. This can be used by
    normal users to clip output. */

/** Constructor.
    \param rect         clipping rectangle
    \param parent       canvas to draw on */
GfxClipRectFilter::GfxClipRectFilter(const GfxRect& rect, GfxCanvas& parent)
    : GfxClipRect(rect, parent)
{ }

/******************************* GfxOffset *******************************/

/** \class GfxOffset
    \brief Offset drawing

    This class provides a filter that offsets all paintings by a
    constant amount. */

/** Constructor.
    \param dx,dy   offset
    \param parent  canvas to draw on */
GfxOffset::GfxOffset(int dx, int dy, GfxCanvas& parent)
    : GfxFilter(parent),
      dx(dx), dy(dy)
{ }

void
GfxOffset::drawHLine(const GfxPoint& pt, int npix, uint32 color, uint8 pat, alpha_type alpha)
{
    parent.drawHLine(pt.movedBy(dx, dy), npix, color, pat, alpha);
}

void
GfxOffset::drawVLine(const GfxPoint& pt, int npix, uint32 color, uint8 pat, alpha_type alpha)
{
    parent.drawVLine(pt.movedBy(dx, dy), npix, color, pat, alpha);
}

void
GfxOffset::drawPixel(const GfxPoint& pt, uint32 color, alpha_type alpha)
{
    parent.drawPixel(GfxPoint(pt.x + dx, pt.y + dy), color, alpha);
}

void
GfxOffset::drawBar(GfxRect rect, uint32 color, uint32 bg,
                   const GfxFillPattern& pat, alpha_type alpha)
{
    rect.moveBy(dx, dy);
    parent.drawBar(rect, color, bg, pat, alpha);
}

void
GfxOffset::blitSurface(const GfxPoint& pt, GfxCanvas& src, GfxRect rect)
{
    parent.blitSurface(GfxPoint(pt.x + dx, pt.y + dy), src, rect);
}

void
GfxOffset::blitPattern(GfxRect rect, const GfxPoint& pt, int bytes_per_line,
                       const uint8* data, uint32 color, uint32 bg,
                       alpha_type alpha)
{
    rect.moveBy(dx, dy);
    parent.blitPattern(rect, GfxPoint(pt.x + dx, pt.y + dy), bytes_per_line,
                       data, color, bg, alpha);
}

GfxRect
GfxOffset::computeClipRect(GfxRect r)
{
    r.moveBy(dx, dy);
    GfxRect q = parent.computeClipRect(r);
    q.moveBy(-dx, -dy);
    return q;
}

bool
GfxOffset::computeOffset(GfxCanvas& other, GfxPoint& pt)
{
    if (&other == this) {
        return true;
    } else {
        if (parent.computeOffset(other, pt)) {
            pt.x += dx;
            pt.y += dy;
            return true;
        } else {
            return false;
        }
    }
}

void
GfxOffset::update(GfxRect r)
{
    r.moveBy(dx, dy);
    parent.update(r);
}

uint32
GfxOffset::getPixel(const GfxPoint& pt)
{
    return parent.getPixel(GfxPoint(pt.x + dx, pt.y + dy));
}
