/**
  *  \file u/t_game_racename.cc
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/racename.h"

#include "u/t_game.h"
#include "u/files.h"
#include "io/memstream.h"
#include "cpluslib/utf8.h"

void
GameRaceNameTestSuite::testRaceNameList()
{
    // Construct a race name list
    MemoryStream ms(string_t((char*)file_racenm, file_racenm_size));
    GRaceNameList rnl(ms);

    // Check it
    TS_ASSERT_EQUALS(rnl.getShortName(1), "The Feds");
    TS_ASSERT_EQUALS(rnl.getFullName(1),  "The Solar Federation");
    TS_ASSERT_EQUALS(rnl.getAdjName(1),   "Fed");

    TS_ASSERT_EQUALS(rnl.getShortName(11), "The Colonies");
    TS_ASSERT_EQUALS(rnl.getFullName(11),  "The Missing Colonies of Man");
    TS_ASSERT_EQUALS(rnl.getAdjName(11),   "Colonial");

    // We cannot sensibly check index 0 and 12, because those are mangled
    // by NLS which is in an indeterminate state here. So we just check
    // that they don't throw or otherwise fail.
    rnl.getShortName(0);
    rnl.getFullName(0);
    rnl.getAdjName(0);
    rnl.getShortName(12);
    rnl.getFullName(12);
    rnl.getAdjName(12);

    // Expansions
    TS_ASSERT_EQUALS(rnl.expandNames("a %-5 ship"), "a Privateer ship");
    TS_ASSERT_EQUALS(rnl.expandNames("attack %5!"), "attack The Privateers!");
    TS_ASSERT_EQUALS(rnl.expandNames("%1..."), "The Feds...");
    TS_ASSERT_EQUALS(rnl.expandNames("%1."), "The Feds.");
    TS_ASSERT_EQUALS(rnl.expandNames("%1"), "The Feds");
    TS_ASSERT_EQUALS(rnl.expandNames("%a..."), "The Rebels...");
    TS_ASSERT_EQUALS(rnl.expandNames("...%b"), "...The Colonies");
    TS_ASSERT_EQUALS(rnl.expandNames("%A..."), "The Rebels...");
    TS_ASSERT_EQUALS(rnl.expandNames("...%B"), "...The Colonies");
    TS_ASSERT_EQUALS(rnl.expandNames("%-A..."), "Rebel...");
    TS_ASSERT_EQUALS(rnl.expandNames("...%-B"), "...Colonial");
    TS_ASSERT_EQUALS(rnl.expandNames("%1%2"), "The FedsThe Lizards");
    TS_ASSERT_EQUALS(rnl.expandNames("a%%b"), "a%b");
    TS_ASSERT_EQUALS(rnl.expandNames("%%"), "%");
    TS_ASSERT_EQUALS(rnl.expandNames("%%1"), "%1");

    // Those are out-of-spec. As of 20110102, '%' quotes, i.e. keeps the
    // offending character (this is to avoid eating partial UTF-8 runes).
    TS_ASSERT_EQUALS(rnl.expandNames("%0"), "0");
    TS_ASSERT_EQUALS(rnl.expandNames("%01"), "01");
    TS_ASSERT_EQUALS(rnl.expandNames("%c"), "c");
    TS_ASSERT_EQUALS(rnl.expandNames("%x"), "x");
    TS_ASSERT_EQUALS(rnl.expandNames("%."), ".");
    TS_ASSERT_EQUALS(rnl.expandNames("%-"), "");
    TS_ASSERT_EQUALS(rnl.expandNames("%-."), ".");
    TS_ASSERT_EQUALS(rnl.expandNames("%"), "");

    // Make sure this doesn't produce error characters by breaking UTF-8
    string_t a = rnl.expandNames("a%\xc3\x80");
    Utf8Reader rdr(a);
    while (rdr.hasMore()) {
        uint16_t ch = rdr.get();
        TS_ASSERT(!Utf8Reader::isErrorCharacter(ch));
    }
}
