/**
  *  \file u/t_io_memdir.cc
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "io/memdir.h"
#include "u/t_io.h"
#include "util/except.h"

// Test basic MemoryDirectory API
void
IoMemdirTestSuite::testBaseAPI()
{
    // Create directory
    Ptr<MemoryDirectory> md = MemoryDirectory::create("tester");
    TS_ASSERT(md != 0);
    TS_ASSERT_EQUALS(md->getName(), "tester");

    // Check some operations on blank FS:
    TS_ASSERT_EQUALS(md->eraseStream("foo"), false);
    TS_ASSERT_EQUALS(md->renameStream("foo", "bar"), false);
    TS_ASSERT(md->getStreamByName("foo", false) == 0);

    // Create a file
    Ptr<MemoryStream> ms = md->getStreamByName("foo", true);
    TS_ASSERT(ms != 0);
    TS_ASSERT_EQUALS(ms->getName(), "tester(foo)");
    TS_ASSERT_EQUALS(ms->getSize(), 0);
    TS_ASSERT_EQUALS(ms->getPos(), 0);
    TS_ASSERT_EQUALS(ms, md->getStreamByName("foo", false));

    // Create another file
    Ptr<MemoryStream> ms2 = md->getStreamByName("bar", true);
    TS_ASSERT(ms2 != 0);
    TS_ASSERT_DIFFERS(ms2, ms);
    TS_ASSERT_EQUALS(ms2->getName(), "tester(bar)");
    TS_ASSERT_EQUALS(ms2->getSize(), 0);
    TS_ASSERT_EQUALS(ms2->getPos(), 0);
    TS_ASSERT_EQUALS(ms2, md->getStreamByName("bar", false));

    // Rename file
    TS_ASSERT_EQUALS(md->renameStream("foo", "bar"), true);
    TS_ASSERT(md->getStreamByName("foo", false) == 0);
    TS_ASSERT(md->getStreamByName("bar", false) == ms);
    TS_ASSERT_EQUALS(ms->getName(), "tester(bar)");

    // Erase file
    TS_ASSERT_EQUALS(md->eraseStream("foo"), false);
    TS_ASSERT_EQUALS(md->eraseStream("bar"), true);
    TS_ASSERT_EQUALS(md->eraseStream("bar"), false);
    TS_ASSERT(md->getStreamByName("foo", false) == 0);
    TS_ASSERT(md->getStreamByName("bar", false) == 0);
}

// Test that we can create and access files using regular APIs
void
IoMemdirTestSuite::testCreation()
{
    char tmp[100];

    // Create directory
    Ptr<Directory> md = MemoryDirectory::create("tester");
    TS_ASSERT_EQUALS(md->getName(), "tester");

    // Open fails for nonexistant file
    TS_ASSERT_EQUALS(md->openFileNT("foo", Stream::C_READ), Ptr<Stream>(0));
    TS_ASSERT_THROWS(md->openFile("foo", Stream::C_READ), FileOpenException);
    TS_ASSERT_EQUALS(md->eraseNT("foo"), false);
    TS_ASSERT_THROWS(md->erase("foo"), FileProblemException);
    TS_ASSERT_DIFFERS(md->getEntryByName("foo"), Ptr<DirectoryEntry>(0));
    TS_ASSERT_THROWS(md->getEntryByName("foo")->renameTo("bar"), FileProblemException);

    // Test creation
    Ptr<Stream> s = md->openFile("foo", Stream::C_CREATE);
    TS_ASSERT(s != 0);
    TS_ASSERT_EQUALS(s->getPos(), 0);
    TS_ASSERT_EQUALS(s->getSize(), 0);
    TS_ASSERT_EQUALS(s->write("hello", 5), 5U);
    TS_ASSERT_EQUALS(s->getPos(), 5);
    TS_ASSERT_EQUALS(s->getSize(), 5);

    // Test re-opening / reading
    Ptr<Stream> s2 = md->openFile("foo", Stream::C_READ);
    TS_ASSERT(s2 != 0);
    TS_ASSERT_EQUALS(s2->getPos(), 0);
    TS_ASSERT_EQUALS(s2->getSize(), 5);
    TS_ASSERT_EQUALS(s2->read(tmp, 10), 5U);
    TS_ASSERT_SAME_DATA(tmp, "hello", 5);
    TS_ASSERT_EQUALS(s2->getPos(), 5);
    TS_ASSERT_EQUALS(s2->getSize(), 5);

    Ptr<Stream> s3 = md->openFile("foo", Stream::C_READ | Stream::C_WRITE);
    TS_ASSERT(s3 != 0);
    TS_ASSERT_EQUALS(s3->getPos(), 0);
    TS_ASSERT_EQUALS(s3->getSize(), 5);
    TS_ASSERT_EQUALS(s3->read(tmp, 3), 3U);
    TS_ASSERT_SAME_DATA(tmp, "hel", 3);
    TS_ASSERT_EQUALS(s3->getPos(), 3);
    TS_ASSERT_EQUALS(s3->getSize(), 5);

    // Update through file pointer 3
    TS_ASSERT_EQUALS(s3->write("p me!", 5), 5U);
    TS_ASSERT_EQUALS(s3->getPos(), 8);
    TS_ASSERT_EQUALS(s3->getSize(), 8);

    // All other file sizes also changed, but file pointers stayed
    TS_ASSERT_EQUALS(s->getSize(), 8);
    TS_ASSERT_EQUALS(s2->getSize(), 8);
    TS_ASSERT_EQUALS(s->getPos(), 5);
    TS_ASSERT_EQUALS(s2->getPos(), 5);

    // Check that we can read the enlarged file through an old handle
    TS_ASSERT_EQUALS(s2->read(tmp, 100), 3U);
    TS_ASSERT_EQUALS(s2->getPos(), 8);
    TS_ASSERT_SAME_DATA(tmp, "me!", 3);

    // Check that creation resets file to zero size
    Ptr<Stream> s4 = md->openFile("foo", Stream::C_CREATE);
    TS_ASSERT_EQUALS(s4->getSize(), 0);
    // Fails, because it cannot return to its position:
    // TS_ASSERT_EQUALS(s3->getSize(), 0);
    TS_ASSERT_EQUALS(s2->getSize(), 0);
    // Fails, because it cannot return to its position:
    // TS_ASSERT_EQUALS(s->getSize(), 0);
    TS_ASSERT_EQUALS(s2->read(tmp, 10), 0U);
}

void
IoMemdirTestSuite::testIteration()
{
    // Create directory
    Ptr<Directory> md = MemoryDirectory::create("tester");
    TS_ASSERT_EQUALS(md->getName(), "tester");

    // Must be initially empty
    Ptr<Enumeration<Ptr<DirectoryEntry> > > e = md->getEntries();
    TS_ASSERT(e != 0);
    TS_ASSERT(!e->hasMoreElements());

    // Create file with some content
    {
        Ptr<Stream> s = md->openFile("a", Stream::C_CREATE);
        TS_ASSERT_EQUALS(s->write("hallo", 5), 5U);
    }

    // Must now contain one item with correct specs
    e = md->getEntries();
    TS_ASSERT(e != 0);
    TS_ASSERT(e->hasMoreElements());
    Ptr<DirectoryEntry> entry = e->getNextElement();
    TS_ASSERT(entry != 0);
    TS_ASSERT_EQUALS(entry->getBaseName(), "a");
    TS_ASSERT_EQUALS(entry->getName(), "tester(a)");
    TS_ASSERT_EQUALS(entry->getSize(), 5);
    TS_ASSERT_EQUALS(entry->getFileType(), int(DirectoryEntry::FT_FILE));

    // Entry must be open-able as file
    {
        Ptr<Stream> s;
        TS_ASSERT_THROWS_NOTHING(s = entry->openFile(Stream::C_READ));
        TS_ASSERT(s != 0);
        TS_ASSERT_EQUALS(s->getSize(), 5);
        TS_ASSERT_EQUALS(s->getName(), "tester(a)");
    }

    // Entry must not be open-able as directory
    TS_ASSERT_THROWS(entry->openDir(), FileProblemException);

    // Entry must be linked to directory
    TS_ASSERT(entry->openParent() == md);

    // Directory must not contain more elements
    TS_ASSERT(!e->hasMoreElements());
}
